import logging
import math
from itertools import product
from typing import List, Tuple

import numpy as np

from ..geometry import Vec3Int
from .view import View


def upsample_cube(cube_buffer: np.ndarray, factors: List[int]) -> np.ndarray:
    ds = cube_buffer.shape
    out_buf = np.zeros(tuple(s * f for s, f in zip(ds, factors)), cube_buffer.dtype)
    for dx in (0, factors[0] - 1):
        for dy in (0, factors[1] - 1):
            for dz in (0, factors[2] - 1):
                out_buf[
                    dx : out_buf.shape[0] : factors[0],
                    dy : out_buf.shape[1] : factors[1],
                    dz : out_buf.shape[2] : factors[2],
                ] = cube_buffer
    return out_buf


def upsample_cube_job(
    args: Tuple[View, View, int],
    mag_factors: List[float],
    buffer_shape: Vec3Int,
) -> None:
    (source_view, target_view, _i) = args

    assert all(
        1 >= f for f in mag_factors
    ), f"mag_factors ({mag_factors}) for upsampling must be smaller than 1"

    try:
        num_channels = target_view.info.num_channels
        target_bbox_in_mag = target_view.bounding_box.in_mag(target_view.mag)
        shape = (num_channels,) + target_bbox_in_mag.size.to_tuple()
        shape_xyz = target_bbox_in_mag.size_xyz
        file_buffer = np.zeros(shape, target_view.get_dtype())

        tiles = product(
            *list(
                [
                    list(range(0, math.ceil(length)))
                    for length in shape_xyz.to_np() / buffer_shape.to_np()
                ]
            )
        )

        for tile in tiles:
            target_offset = Vec3Int(tile) * buffer_shape
            source_offset = target_offset * source_view.mag
            source_size = source_view.bounding_box.size_xyz
            source_size = (buffer_shape * source_view.mag).pairmin(
                source_size - source_offset
            )

            bbox = source_view.bounding_box.offset(source_offset).with_size_xyz(
                source_size
            )
            cube_buffer_channels = source_view.read_xyz(
                absolute_bounding_box=bbox,
            )

            for channel_index in range(num_channels):
                cube_buffer = cube_buffer_channels[channel_index]

                if not np.all(cube_buffer == 0):
                    # Upsample the buffer
                    inverse_factors = [int(1 / f) for f in mag_factors]
                    data_cube = upsample_cube(cube_buffer, inverse_factors)

                    buffer_bbox = target_view.bounding_box.with_topleft_xyz(
                        target_offset
                    ).with_size_xyz(data_cube.shape)
                    data_cube = buffer_bbox.xyz_array_to_bbox_shape(data_cube)
                    file_buffer[(channel_index,) + buffer_bbox.to_slices_xyz()] = (
                        data_cube
                    )

        target_view.write(file_buffer, absolute_bounding_box=target_view.bounding_box)

    except Exception as exc:
        logging.error(
            f"Upsampling of target {target_view.bounding_box} failed with {exc}"
        )
        raise exc



from collections import defaultdict
from itertools import product
from typing import (
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
    cast,
)

import attr
import numpy as np

from .mag import Mag
from .vec3_int import Vec3Int, Vec3IntLike
from .vec_int import VecInt, VecIntLike

_DEFAULT_BBOX_NAME = "Unnamed Bounding Box"

_T = TypeVar("_T", bound="NDBoundingBox")


def str_tpl(str_list: Iterable[str]) -> Tuple[str, ...]:
    # Fix for mypy bug https://github.com/python/mypy/issues/5313.
    # Solution based on other issue for the same bug: https://github.com/python/mypy/issues/8389.
    return tuple(str_list)


def int_tpl(vec_int_like: VecIntLike) -> VecInt:
    return VecInt(
        vec_int_like, axes=(f"unset_{i}" for i in range(len(list(vec_int_like))))
    )


@attr.frozen
class NDBoundingBox:
    """
    The NDBoundingBox class is a generalized version of the 3-dimensional BoundingBox class. It is designed to represent bounding boxes in any number of dimensions.

    The bounding box is characterized by its top-left corner, the size of the box, the names of the axes for each dimension, and the index (or order) of the axes. Each axis must have a unique index, starting from 1 (index 0 is reserved for channel information).

    The top-left coordinate is inclusive, while the bottom-right coordinate is exclusive.

    Here's a brief example of how to use it:

    ```python

    # Create a 2D bounding box
    bbox_1 = NDBoundingBox(
        top_left=(0, 0),
        size=(100, 100),
        axes=("x", "y"),
        index=(1,2)
    )

    # Create a 4D bounding box
    bbox_2 = NDBoundingBox(
        top_left=(75, 75, 75, 0),
        size=(100, 100, 100, 20),
        axes=("x", "y", "z", "t"),
        index=(2,3,4,1)
    )
    ```
    """

    topleft: VecInt = attr.field(converter=int_tpl)
    size: VecInt = attr.field(converter=int_tpl)
    axes: Tuple[str, ...] = attr.field(converter=str_tpl)
    index: VecInt = attr.field(converter=int_tpl)
    bottomright: VecInt = attr.field(init=False)
    name: Optional[str] = _DEFAULT_BBOX_NAME
    is_visible: bool = True
    color: Optional[Tuple[float, float, float, float]] = None

    def __attrs_post_init__(self) -> None:
        assert (
            len(self.topleft) == len(self.size) == len(self.axes) == len(self.index)
        ), (
            f"The dimensions of topleft, size, axes and index ({len(self.topleft)}, "
            + f"{len(self.size)}, {len(self.axes)} and {len(self.index)}) do not match."
        )
        assert 0 not in self.index, "Index 0 is reserved for channels."

        # Convert the delivered tuples to VecInts
        object.__setattr__(self, "topleft", VecInt(self.topleft, axes=self.axes))
        object.__setattr__(self, "size", VecInt(self.size, axes=self.axes))
        object.__setattr__(self, "index", VecInt(self.index, axes=self.axes))

        if not self._is_sorted():
            self._sort_positions_of_axes()

        if not self.size.is_positive():
            # Flip the size in negative dimensions, so that the topleft is smaller than bottomright.
            # E.g. BoundingBox((10, 10, 10), (-5, 5, 5)) -> BoundingBox((5, 10, 10), (5, 5, 5)).
            negative_size = tuple(min(0, value) for value in self.size)
            new_topleft = tuple(
                val1 + val2 for val1, val2 in zip(self.topleft, negative_size)
            )
            new_size = (abs(value) for value in self.size)
            object.__setattr__(self, "topleft", VecInt(new_topleft, axes=self.axes))
            object.__setattr__(self, "size", VecInt(new_size, axes=self.axes))

        # Compute bottomright to avoid that it's recomputed every time
        # it is needed.
        object.__setattr__(
            self,
            "bottomright",
            self.topleft + self.size,
        )

    def _sort_positions_of_axes(self) -> None:
        # Bring topleft and size in required order
        # defined in axisOrder and index of additionalAxes

        size, topleft, axes, index = zip(
            *sorted(
                zip(self.size, self.topleft, self.axes, self.index), key=lambda x: x[3]
            )
        )
        object.__setattr__(self, "size", VecInt(size, axes=axes))
        object.__setattr__(self, "topleft", VecInt(topleft, axes=axes))
        object.__setattr__(self, "axes", axes)
        object.__setattr__(self, "index", VecInt(index, axes=axes))

    def _is_sorted(self) -> bool:
        return all(self.index[i - 1] < self.index[i] for i in range(1, len(self.index)))

    def with_name(self: _T, name: Optional[str]) -> _T:
        """
        Returns a new instance of `NDBoundingBox` with the specified name.

        Args:
        - name (Optional[str]): The name to assign to the new `NDBoundingBox` instance.

        Returns:
        - NDBoundingBox: A new instance of `NDBoundingBox` with the specified name.
        """
        return attr.evolve(self, name=name)

    def with_topleft(self: _T, new_topleft: VecIntLike) -> _T:
        """
        Returns a new NDBoundingBox object with the specified top left coordinates.

        Args:
        - new_topleft (VecIntLike): The new top left coordinates for the bounding box.

        Returns:
        - NDBoundingBox: A new NDBoundingBox object with the updated top left coordinates.
        """
        return attr.evolve(self, topleft=VecInt(new_topleft, axes=self.axes))

    def with_size(self: _T, new_size: VecIntLike) -> _T:
        """
        Returns a new NDBoundingBox object with the specified size.

        Args:
        - new_size (VecIntLike): The new size of the bounding box. Can be a VecInt or any object that can be converted to a VecInt.

        Returns:
        - A new NDBoundingBox object with the specified size.
        """
        return attr.evolve(self, size=VecInt(new_size, axes=self.axes))

    def with_index(self: _T, new_index: VecIntLike) -> _T:
        """
        Returns a new NDBoundingBox object with the specified index.

        Args:
        - new_index (VecIntLike): The new axis order for the bounding box.

        Returns:
        - NDBoundingBox: A new NDBoundingBox object with the updated index.
        """
        axes, _ = zip(*sorted(zip(self.axes, new_index), key=lambda x: x[1]))
        return attr.evolve(self, index=VecInt(new_index, axes=axes))

    def with_bottomright(self: _T, new_bottomright: VecIntLike) -> _T:
        """
        Returns a new NDBoundingBox with an updated bottomright value.

        Args:
        - new_bottomright (VecIntLike): The new bottom right corner coordinates.

        Returns:
        - NDBoundingBox: A new NDBoundingBox object with the updated bottom right corner.
        """
        new_size = VecInt(new_bottomright, axes=self.axes) - self.topleft

        return self.with_size(new_size)

    def with_is_visible(self: _T, is_visible: bool) -> _T:
        """
        Returns a new NDBoundingBox object with the specified visibility.

        Args:
        - is_visible (bool): The visibility value to set.

        Returns:
        - NDBoundingBox: A new NDBoundingBox object with the updated visibility value.
        """
        return attr.evolve(self, is_visible=is_visible)

    def with_color(self: _T, color: Optional[Tuple[float, float, float, float]]) -> _T:
        """
        Returns a new instance of NDBoundingBox with the specified color.

        Args:
        - color (Optional[Tuple[float, float, float, float]]): The color to set for the bounding box.
            The color should be specified as a tuple of four floats representing RGBA values.

        Returns:
        - NDBoundingBox: A new instance of NDBoundingBox with the specified color.
        """
        return attr.evolve(self, color=color)

    def with_bounds(
        self: _T, axis: str, new_topleft: Optional[int], new_size: Optional[int]
    ) -> _T:
        """
        Returns a new NDBoundingBox object with updated bounds along the specified axis.

        Args:
        - axis (str): The name of the axis to update.
        - new_topleft (Optional[int]): The new value for the top-left coordinate along the specified axis.
        - new_size (Optional[int]): The new size along the specified axis.

        Returns:
        - NDBoundingBox: A new NDBoundingBox object with updated bounds.

        Raises:
        - ValueError: If the given axis name does not exist.

        """
        try:
            index = self.axes.index(axis)
        except ValueError as err:
            raise ValueError("The given axis name does not exist.") from err

        _new_topleft = (
            self.topleft.with_replaced(index, new_topleft)
            if new_topleft is not None
            else self.topleft
        )
        _new_size = (
            self.size.with_replaced(index, new_size)
            if new_size is not None
            else self.size
        )

        return attr.evolve(self, topleft=_new_topleft, size=_new_size)

    def get_bounds(self, axis: str) -> Tuple[int, int]:
        """
        Returns the bounds of the given axis.

        Args:
        - axis (str): The name of the axis to get the bounds for.

        Returns:
        - Tuple[int, int]: A tuple containing the top-left and bottom-right coordinates along the specified axis.
        """
        try:
            index = self.axes.index(axis)
        except ValueError as err:
            raise ValueError("The given axis name does not exist.") from err

        return (self.topleft[index], self.topleft[index] + self.size[index])

    @classmethod
    def group_boxes_with_aligned_mag(
        cls, bounding_boxes: Iterable["NDBoundingBox"], aligning_mag: Mag
    ) -> Dict["NDBoundingBox", List["NDBoundingBox"]]:
        """
        Groups the given BoundingBox instances by aligning each
        bbox to the given mag and using that as the key.
        For example, bounding boxes of size 256**3 could be grouped
        into the corresponding 1024**3 chunks to which they belong
        by using aligning_mag = Mag(1024).
        """

        chunks_with_bboxes = defaultdict(list)
        for bbox in bounding_boxes:
            chunk_key = bbox.align_with_mag(aligning_mag, ceil=True)
            chunks_with_bboxes[chunk_key].append(bbox)

        return chunks_with_bboxes

    @classmethod
    def from_wkw_dict(cls, bbox: Dict) -> "NDBoundingBox":
        """
        Create an instance of NDBoundingBox from a dictionary representation.

        Args:
        - bbox (Dict): The dictionary representation of the bounding box.

        Returns:
        - NDBoundingBox: An instance of NDBoundingBox.

        Raises:
        - AssertionError: If additionalAxes are present but axisOrder is not provided.
        """

        topleft: Tuple[int, ...] = bbox["topLeft"]
        size: Tuple[int, ...] = (bbox["width"], bbox["height"], bbox["depth"])
        axes: Tuple[str, ...] = ("x", "y", "z")
        index: Tuple[int, ...] = (1, 2, 3)

        if "axisOrder" in bbox:
            axes = tuple(bbox["axisOrder"].keys())
            index = tuple(bbox["axisOrder"][axis] for axis in axes)

            if "additionalAxes" in bbox:
                assert (
                    "axisOrder" in bbox
                ), "If there are additionalAxes an axisOrder needs to be provided."
                for axis in bbox["additionalAxes"]:
                    topleft += (axis["bounds"][0],)
                    size += (axis["bounds"][1] - axis["bounds"][0],)
                    axes += (axis["name"],)
                    index += (axis["index"],)

        return cls(
            topleft=VecInt(topleft, axes=axes),
            size=VecInt(size, axes=axes),
            axes=axes,
            index=VecInt(index, axes=axes),
        )

    def to_wkw_dict(self) -> dict:
        """
        Converts the bounding box object to a json dictionary.

        Returns:
        - dict: A json dictionary representing the bounding box.
        """
        topleft = [None, None, None]
        width, height, depth = None, None, None
        additional_axes = []
        for i, axis in enumerate(self.axes):
            if axis == "x":
                topleft[0] = self.topleft[i]
                width = self.size[i]
            elif axis == "y":
                topleft[1] = self.topleft[i]
                height = self.size[i]
            elif axis == "z":
                topleft[2] = self.topleft[i]
                depth = self.size[i]
            else:
                additional_axes.append(
                    {
                        "name": axis,
                        "bounds": [self.topleft[i], self.bottomright[i]],
                        "index": self.index[i],
                    }
                )
        if additional_axes:
            return {
                "topLeft": topleft,
                "width": width,
                "height": height,
                "depth": depth,
                "additionalAxes": additional_axes,
            }
        return {
            "topLeft": topleft,
            "width": width,
            "height": height,
            "depth": depth,
        }

    def to_config_dict(self) -> dict:
        """
        Returns a dictionary representation of the bounding box.

        Returns:
        - dict: A dictionary representation of the bounding box.
        """
        return {
            "topleft": self.topleft.to_list(),
            "size": self.size.to_list(),
            "axes": self.axes,
        }

    def to_checkpoint_name(self) -> str:
        """
        Returns a string representation of the bounding box that can be used as a checkpoint name.

        Returns:
        - str: A string representation of the bounding box.
        """
        return f"{'_'.join(str(element) for element in self.topleft)}_{'_'.join(str(element) for element in self.size)}"

    def __repr__(self) -> str:
        return f"NDBoundingBox(topleft={self.topleft.to_tuple()}, size={self.size.to_tuple()}, axes={self.axes})"

    def __str__(self) -> str:
        return self.__repr__()

    def __eq__(self, other: object) -> bool:
        if isinstance(other, NDBoundingBox):
            self._check_compatibility(other)
            return self.topleft == other.topleft and self.size == other.size

        raise NotImplementedError()

    def __len__(self) -> int:
        return len(self.axes)

    def get_shape(self, axis_name: str) -> int:
        """
        Returns the size of the bounding box along the specified axis.

        Args:
        - axis_name (str): The name of the axis to get the size for.

        Returns:
        - int: The size of the bounding box along the specified axis.
        """
        try:
            index = self.axes.index(axis_name)
            return self.size[index]
        except ValueError as err:
            raise ValueError(
                f"Axis {axis_name} doesn't exist in NDBoundingBox."
            ) from err

    def _get_attr_xyz(self, attr_name: str) -> Vec3Int:
        axes = ("x", "y", "z")
        attr_3d = []

        for axis in axes:
            index = self.axes.index(axis)
            attr_3d.append(getattr(self, attr_name)[index])

        return Vec3Int(attr_3d)

    def _get_attr_with_replaced_xyz(self, attr_name: str, xyz: Vec3IntLike) -> VecInt:
        value = Vec3Int(xyz)
        axes = ("x", "y", "z")
        modified_attr = getattr(self, attr_name).to_list()

        for i, axis in enumerate(axes):
            index = self.axes.index(axis)
            modified_attr[index] = value[i]

        return VecInt(modified_attr, axes=self.axes)

    @property
    def topleft_xyz(self) -> Vec3Int:
        """The topleft corner of the bounding box regarding only x, y and z axis."""

        return self._get_attr_xyz("topleft")

    @property
    def size_xyz(self) -> Vec3Int:
        """The size of the bounding box regarding only x, y and z axis."""

        return self._get_attr_xyz("size")

    @property
    def bottomright_xyz(self) -> Vec3Int:
        """The bottomright corner of the bounding box regarding only x, y and z axis."""

        return self._get_attr_xyz("bottomright")

    @property
    def index_xyz(self) -> Vec3Int:
        """The index of x, y and z axis within the bounding box."""

        return self._get_attr_xyz("index")

    def with_topleft_xyz(self: _T, new_xyz: Vec3IntLike) -> _T:
        """
        Returns a new NDBoundingBox object with changed x, y and z coordinates of the topleft corner.

        Args:
        - new_xyz (Vec3IntLike): The new x, y and z coordinates for the topleft corner.

        Returns:
        - NDBoundingBox: A new NDBoundingBox object with the updated x, y and z coordinates of the topleft corner.
        """
        new_topleft = self._get_attr_with_replaced_xyz("topleft", new_xyz)

        return self.with_topleft(new_topleft)

    def with_size_xyz(self: _T, new_xyz: Vec3IntLike) -> _T:
        """
        Returns a new NDBoundingBox object with changed x, y and z size.

        Args:
        - new_xyz (Vec3IntLike): The new x, y and z size for the bounding box.

        Returns:
        - NDBoundingBox: A new NDBoundingBox object with the updated x, y and z size.
        """
        new_size = self._get_attr_with_replaced_xyz("size", new_xyz)

        return self.with_size(new_size)

    def with_bottomright_xyz(self: _T, new_xyz: Vec3IntLike) -> _T:
        """
        Returns a new NDBoundingBox object with changed x, y and z coordinates of the bottomright corner.

        Args:
        - new_xyz (Vec3IntLike): The new x, y and z coordinates for the bottomright corner.

        Returns:
        - NDBoundingBox: A new NDBoundingBox object with the updated x, y and z coordinates of the bottomright corner.
        """
        new_bottomright = self._get_attr_with_replaced_xyz("bottomright", new_xyz)

        return self.with_bottomright(new_bottomright)

    def with_index_xyz(self: _T, new_xyz: Vec3IntLike) -> _T:
        """
        Returns a new NDBoundingBox object with changed x, y and z index.

        Args:
        - new_xyz (Vec3IntLike): The new x, y and z index for the bounding box.

        Returns:
        - NDBoundingBox: A new NDBoundingBox object with the updated x, y and z index.
        """
        new_index = self._get_attr_with_replaced_xyz("index", new_xyz)

        return self.with_index(new_index)

    def _check_compatibility(self, other: "NDBoundingBox") -> None:
        """Checks if two bounding boxes are comparable. To be comparable they need the same number of axes, with same names and same order."""

        if self.axes != other.axes:
            raise ValueError(
                f"Operation with two bboxes is only possible if they have the same axes and axes order. {self.axes} != {other.axes}"
            )

    def padded_with_margins(
        self: _T, margins_left: Vec3IntLike, margins_right: Optional[Vec3IntLike] = None
    ) -> _T:
        if margins_right is None:
            margins_right = margins_left

        margins_left = Vec3Int(margins_left)
        margins_right = Vec3Int(margins_right)

        return self.with_topleft_xyz(self.topleft_xyz - margins_left).with_size_xyz(
            self.size_xyz + (margins_left + margins_right)
        )

    def intersected_with(self: _T, other: _T, dont_assert: bool = False) -> _T:
        """
        Returns the intersection of two bounding boxes.

        If dont_assert is set to False, this method may return empty bounding boxes (size == (0, 0, 0))

        Args:
        - other (NDBoundingBox): The other bounding box to intersect with.
        - dont_assert (bool): If True, the method may return empty bounding boxes.

        Returns:
        - NDBoundingBox: The intersection of the two bounding boxes.
        """

        self._check_compatibility(other)
        topleft = self.topleft.pairmax(other.topleft)
        bottomright = self.bottomright.pairmin(other.bottomright)
        size = (bottomright - topleft).pairmax(VecInt.zeros(self.axes))

        intersection = attr.evolve(self, topleft=topleft, size=size)

        if not dont_assert:
            assert (
                not intersection.is_empty()
            ), f"No intersection between bounding boxes {self} and {other}."

        return intersection

    def extended_by(self: _T, other: _T) -> _T:
        """
        Returns the smallest bounding box that contains both bounding boxes.

        Args:
        - other (NDBoundingBox): The other bounding box to extend with.

        Returns:
        - NDBoundingBox: The smallest bounding box that contains both bounding boxes.
        """
        self._check_compatibility(other)
        if self.is_empty():
            return other
        if other.is_empty():
            return self

        topleft = self.topleft.pairmin(other.topleft)
        bottomright = self.bottomright.pairmax(other.bottomright)
        size = bottomright - topleft

        return attr.evolve(self, topleft=topleft, size=size)

    def is_empty(self) -> bool:
        """
        Boolean check whether the boundung box is empty.

        Returns:
        - bool: True if the bounding box is empty, False otherwise.
        """
        return not self.size.is_positive(strictly_positive=True)

    def in_mag(self: _T, mag: Mag) -> _T:
        """
        Returns the bounding box in the given mag.

        Args:
        - mag (Mag): The magnification to convert the bounding box to.

        Returns:
        - NDBoundingBox: The bounding box in the given magnification.
        """
        mag_vec = mag.to_vec3_int()

        assert (
            self.topleft_xyz % mag_vec == Vec3Int.zeros()
        ), f"topleft {self.topleft} is not aligned with the mag {mag}. Use BoundingBox.align_with_mag()."
        assert (
            self.bottomright_xyz % mag_vec == Vec3Int.zeros()
        ), f"bottomright {self.bottomright} is not aligned with the mag {mag}. Use BoundingBox.align_with_mag()."

        return self.with_topleft_xyz(self.topleft_xyz // mag_vec).with_size_xyz(
            self.size_xyz // mag_vec
        )

    def from_mag_to_mag1(self: _T, from_mag: Mag) -> _T:
        """
        Returns the bounging box in the finest magnification (Mag(1)).

        Args:
        - from_mag (Mag): The current magnification of the bounding box.

        Returns:
        - NDBoundingBox: The bounding box in the given magnification.
        """
        mag_vec = from_mag.to_vec3_int()

        return self.with_topleft_xyz(self.topleft_xyz * mag_vec).with_size_xyz(
            self.size_xyz * mag_vec
        )

    def _align_with_mag_slow(self: _T, mag: Mag, ceil: bool = False) -> _T:
        """Rounds the bounding box, so that both topleft and bottomright are divisible by mag.

        :argument ceil: If true, the bounding box is enlarged when necessary. If false, it's shrunk when necessary.
        """
        np_mag = mag.to_np()

        align = (  # noqa E731
            lambda point, round_fn: round_fn(point.to_np() / np_mag).astype(int)
            * np_mag
        )

        if ceil:
            topleft = align(self.topleft, np.floor)
            bottomright = align(self.bottomright, np.ceil)
        else:
            topleft = align(self.topleft, np.ceil)
            bottomright = align(self.bottomright, np.floor)
        return attr.evolve(self, topleft=topleft, size=bottomright - topleft)

    def align_with_mag(self: _T, mag: Union[Mag, Vec3Int], ceil: bool = False) -> _T:
        """
        Rounds the bounding box, so that both topleft and bottomright are divisible by mag.

        Args:
        - mag (Union[Mag, Vec3Int]): The magnification to align the bounding box to.
        - ceil (bool): If True, the bounding box is enlarged when necessary. If False, it's shrunk when necessary.

        Returns:
        - NDBoundingBox: The aligned bounding box.
        """
        # This does the same as _align_with_mag_slow, which is more readable.
        # Same behavior is asserted in test_align_with_mag_against_numpy_implementation
        mag_vec = mag.to_vec3_int() if isinstance(mag, Mag) else mag
        topleft = self.topleft_xyz
        bottomright = self.bottomright_xyz
        roundup = topleft if ceil else bottomright
        rounddown = bottomright if ceil else topleft
        margin_to_roundup = roundup % mag_vec
        aligned_roundup = roundup - margin_to_roundup
        margin_to_rounddown = (mag_vec - (rounddown % mag_vec)) % mag_vec
        aligned_rounddown = rounddown + margin_to_rounddown
        if ceil:
            return self.with_topleft_xyz(aligned_roundup).with_size_xyz(
                aligned_rounddown - aligned_roundup
            )
        else:
            return self.with_topleft_xyz(aligned_rounddown).with_size_xyz(
                aligned_roundup - aligned_rounddown
            )

    def contains(self, coord: VecIntLike) -> bool:
        """
        Check whether a point is inside of the bounding box.
        Note that the point may have float coordinates in the ndarray case

        Args:
        - coord (VecIntLike): The coordinates to check.

        Returns:
        - bool: True if the point is inside of the bounding box, False otherwise.
        """

        if isinstance(coord, np.ndarray):
            assert (
                coord.shape == (len(self.size),)
            ), f"Numpy array BoundingBox.contains must have shape ({len(self.size)},), got {coord.shape}."
            return cast(
                bool,
                np.all(coord >= self.topleft) and np.all(coord < self.bottomright),
            )
        else:
            # In earlier versions, we simply converted to ndarray to have
            # a unified calculation here, but this turned out to be a performance bottleneck.
            # Therefore, the contains-check is performed on the tuple here.
            coord = VecInt(coord, axes=self.axes)
            return all(
                self.topleft[i] <= coord[i] < self.bottomright[i]
                for i in range(len(self.axes))
            )

    def contains_bbox(self: _T, inner_bbox: _T) -> bool:
        """
        Check whether a bounding box is completely inside of the bounding box.

        Args:
        - inner_bbox (NDBoundingBox): The bounding box to check.

        Returns:
        - bool: True if the bounding box is completely inside of the bounding box, False otherwise.
        """
        self._check_compatibility(inner_bbox)
        return inner_bbox.intersected_with(self, dont_assert=True) == inner_bbox

    def chunk(
        self: _T,
        chunk_shape: VecIntLike,
        chunk_border_alignments: Optional[VecIntLike] = None,
    ) -> Generator[_T, None, None]:
        """
        Decompose the bounding box into smaller chunks of size `chunk_shape`.

        Chunks at the border of the bounding box might be smaller than chunk_shape.
        If `chunk_border_alignment` is set, all border coordinates
        *between two chunks* will be divisible by that value.

        Args:
        - chunk_shape (VecIntLike): The size of the chunks to generate.
        - chunk_border_alignments (Optional[VecIntLike]): The alignment of the chunk borders.


        Yields:
        - Generator[NDBoundingBox]: A generator of the chunks.
        """

        start = self.topleft.to_np()
        try:
            # If a 3D chunk_shape is given it is assumed that iteration over xyz is
            # intended. Therefore NDBoundingBoxes are generated that have a shape of
            # x: chunk_shape.x, y: chunk_shape.y, z: chunk_shape.z and 1 for all other
            # axes.
            chunk_shape = Vec3Int(chunk_shape)

            chunk_shape = (
                self.with_size(VecInt.ones(self.axes))
                .with_size_xyz(chunk_shape)
                .size.to_np()
            )
        except AssertionError:
            chunk_shape = VecInt(chunk_shape, axes=self.axes).to_np()

        start_adjust = VecInt.zeros(self.axes).to_np()
        if chunk_border_alignments is not None:
            try:
                chunk_border_alignments = Vec3Int(chunk_border_alignments)

                chunk_border_alignments = (
                    self.with_size(VecInt.ones(self.axes))
                    .with_size_xyz(chunk_border_alignments)
                    .size.to_np()
                )
            except AssertionError:
                chunk_border_alignments = VecInt(
                    chunk_border_alignments, axes=self.axes
                ).to_np()

            assert np.all(
                chunk_shape % chunk_border_alignments == 0
            ), f"{chunk_shape} not divisible by {chunk_border_alignments}"

            # Move the start to be aligned correctly. This doesn't actually change
            # the start of the first chunk, because we'll intersect with `self`,
            # but it'll lead to all chunk borders being aligned correctly.
            start_adjust = start % chunk_border_alignments
        for coordinates in product(
            *[
                range(
                    start[i] - start_adjust[i], start[i] + self.size[i], chunk_shape[i]
                )
                for i in range(len(self.axes))
            ]
        ):
            yield self.intersected_with(
                self.__class__(
                    topleft=VecInt(coordinates, axes=self.axes),
                    size=VecInt(chunk_shape, axes=self.axes),
                    axes=self.axes,
                    index=self.index,
                )
            )

    def volume(self) -> int:
        """
        Returns the volume of the bounding box.
        """
        return self.size.prod()

    def slice_array(self, array: np.ndarray) -> np.ndarray:
        """
        Returns a slice of the given array that corresponds to the bounding box.
        """
        return array[self.to_slices()]

    def xyz_array_to_bbox_shape(self, data: np.ndarray) -> np.ndarray:
        """
        Transforms data array with xyz axes to the shape of the bounding box.
        This is only possible for bboxes that are flat in all dimensions except xyz.
        """
        assert all(
            size == 1 for size, axis in zip(self.size, self.axes) if axis not in "xyz"
        ), "The view's bounding box must be flat in all dimensions except xyz."
        data = np.expand_dims(data, axis=tuple(range(3, len(self))))
        return np.moveaxis(
            data,
            [0, 1, 2],
            (
                self.axes.index("x"),
                self.axes.index("y"),
                self.axes.index("z"),
            ),
        )

    def to_slices(self) -> Tuple[slice, ...]:
        """
        Returns a tuple of slices that corresponds to the bounding box.
        """
        return tuple(
            slice(topleft, topleft + size)
            for topleft, size in zip(self.topleft, self.size)
        )

    def to_slices_xyz(self) -> Tuple[slice, ...]:
        """
        Returns a tuple of slices that corresponds to the bounding box in x, y, z and leaves all other axes one dimensional without offset.
        """
        assert all(
            size == 1 for size, axis in zip(self.size, self.axes) if axis not in "xyz"
        ), "The view's bounding box must be flat in all dimensions except xyz."
        return (
            NDBoundingBox(VecInt.zeros(self.axes), self.size, self.axes, self.index)
            .with_topleft_xyz(self.topleft_xyz)
            .to_slices()
        )

    def offset(self: _T, vector: VecIntLike) -> _T:
        """
        Returns a new NDBoundingBox object with the specified offset.

        Args:
        - vector (VecIntLike): The offset to apply to the bounding box.

        Returns:
        - NDBoundingBox: A new NDBoundingBox object with the specified offset.
        """
        try:
            return self.with_topleft_xyz(self.topleft_xyz + Vec3Int(vector))
        except AssertionError:
            return self.with_topleft(self.topleft + VecInt(vector, axes=self.axes))



import random
from pathlib import Path

import numpy as np

from webknossos.dataset.layer_categories import LayerCategoryType
from webknossos.dataset.mag_view import MagView
from webknossos.geometry.vec3_int import Vec3Int

NUM_SAMPLES = 20
THRESHOLD = (
    1 / 200
)  # more unique values per voxel than this value means color, less segmentation
SAMPLE_SIZE = Vec3Int(16, 16, 16)
MAX_FAILS = 200
MIN_INSPECTED_VOXELS = 1000


def guess_if_segmentation_path(filepath: Path) -> bool:
    lowercase_filepath = str(filepath).lower()
    return any(i in lowercase_filepath for i in ["segmentation", "labels"])


def guess_category_from_view(view: MagView) -> LayerCategoryType:
    if view.layer.num_channels > 1 or sample_distinct_values_per_vx(view) > THRESHOLD:
        return "color"
    return "segmentation"


def sample_distinct_values_per_vx(view: MagView) -> float:
    sample_size_for_view = view.bounding_box.size_xyz.pairmin(SAMPLE_SIZE * view.mag)
    min_offset = view.bounding_box.topleft_xyz
    max_offset = view.bounding_box.bottomright_xyz - sample_size_for_view

    distinct_color_values = 0
    valid_sample_count = 0
    inspected_voxel_count = 0
    invalid_sample_count = 0

    while valid_sample_count < NUM_SAMPLES:
        if invalid_sample_count > MAX_FAILS:
            break
        offset = Vec3Int(
            random.randint(min_offset.x, max_offset.x),
            random.randint(min_offset.y, max_offset.y),
            random.randint(min_offset.z, max_offset.z),
        )
        bbox_to_read = view.bounding_box.with_topleft_xyz(offset).with_size_xyz(
            sample_size_for_view
        )

        data = view.read(absolute_bounding_box=bbox_to_read)
        # The heuristic should avoid checking "empty" areas.
        # As empty, we consider areas that contain only zeros or max values.
        # These values are removed from the data before calculating the distinct values.
        data = data[data != 0]
        try:
            data = data[data != np.iinfo(data.dtype).max]
        except Exception:  # pylint: disable=bare-except
            pass  # does not work or make sense for float data types

        distinct_color_values_in_sample = np.unique(data)

        if len(distinct_color_values_in_sample) == 0:
            invalid_sample_count += 1
        else:
            distinct_color_values += len(distinct_color_values_in_sample)
            valid_sample_count += 1
            inspected_voxel_count += data.size

    if inspected_voxel_count < MIN_INSPECTED_VOXELS:
        raise RuntimeError(
            f"Failed to find enough valid samples (saw {inspected_voxel_count} voxels)."
        )

    return distinct_color_values / inspected_voxel_count



from enum import Enum


class LengthUnit(Enum):
    YOCTOMETER = "yoctometer"
    ZEPTOMETER = "zeptometer"
    ATTOMETER = "attometer"
    FEMTOMETER = "femtometer"
    PICOMETER = "picometer"
    NANOMETER = "nanometer"
    MICROMETER = "micrometer"
    MILLIMETER = "millimeter"
    CENTIMETER = "centimeter"
    DECIMETER = "decimeter"
    METER = "meter"
    HECTOMETER = "hectometer"
    KILOMETER = "kilometer"
    MEGAMETER = "megameter"
    GIGAMETER = "gigameter"
    TERAMETER = "terameter"
    PETAMETER = "petameter"
    EXAMETER = "exameter"
    ZETTAMETER = "zettameter"
    YOTTAMETER = "yottameter"
    ANGSTROM = "angstrom"
    INCH = "inch"
    FOOT = "foot"
    YARD = "yard"
    MILE = "mile"
    PARSEC = "parsec"


_STR_TO_UNIT_MAP = {
    "ym": LengthUnit.YOCTOMETER,
    "yoctometer": LengthUnit.YOCTOMETER,
    "zm": LengthUnit.ZEPTOMETER,
    "zeptometer": LengthUnit.ZEPTOMETER,
    "am": LengthUnit.ATTOMETER,
    "attometer": LengthUnit.ATTOMETER,
    "fm": LengthUnit.FEMTOMETER,
    "femtometer": LengthUnit.FEMTOMETER,
    "pm": LengthUnit.PICOMETER,
    "picometer": LengthUnit.PICOMETER,
    "nm": LengthUnit.NANOMETER,
    "nanometer": LengthUnit.NANOMETER,
    "µm": LengthUnit.MICROMETER,
    "micrometer": LengthUnit.MICROMETER,
    "mm": LengthUnit.MILLIMETER,
    "millimeter": LengthUnit.MILLIMETER,
    "cm": LengthUnit.CENTIMETER,
    "centimeter": LengthUnit.CENTIMETER,
    "dm": LengthUnit.DECIMETER,
    "decimeter": LengthUnit.DECIMETER,
    "m": LengthUnit.METER,
    "meter": LengthUnit.METER,
    "hm": LengthUnit.HECTOMETER,
    "hectometer": LengthUnit.HECTOMETER,
    "km": LengthUnit.KILOMETER,
    "kilometer": LengthUnit.KILOMETER,
    "Mm": LengthUnit.MEGAMETER,
    "megameter": LengthUnit.MEGAMETER,
    "Gm": LengthUnit.GIGAMETER,
    "gigameter": LengthUnit.GIGAMETER,
    "Tm": LengthUnit.TERAMETER,
    "terameter": LengthUnit.TERAMETER,
    "Pm": LengthUnit.PETAMETER,
    "petameter": LengthUnit.PETAMETER,
    "Em": LengthUnit.EXAMETER,
    "exameter": LengthUnit.EXAMETER,
    "Zm": LengthUnit.ZETTAMETER,
    "zettameter": LengthUnit.ZETTAMETER,
    "Ym": LengthUnit.YOTTAMETER,
    "yottameter": LengthUnit.YOTTAMETER,
    "Å": LengthUnit.ANGSTROM,
    "angstrom": LengthUnit.ANGSTROM,
    "in": LengthUnit.INCH,
    "inch": LengthUnit.INCH,
    "ft": LengthUnit.FOOT,
    "foot": LengthUnit.FOOT,
    "yd": LengthUnit.YARD,
    "yard": LengthUnit.YARD,
    "mi": LengthUnit.MILE,
    "mile": LengthUnit.MILE,
    "pc": LengthUnit.PARSEC,
    "parsec": LengthUnit.PARSEC,
}

_LENGTH_UNIT_TO_NANOMETER = {
    LengthUnit.YOCTOMETER: 1e-15,
    LengthUnit.ZEPTOMETER: 1e-12,
    LengthUnit.ATTOMETER: 1e-9,
    LengthUnit.FEMTOMETER: 1e-6,
    LengthUnit.PICOMETER: 1e-3,
    LengthUnit.NANOMETER: 1.0,
    LengthUnit.MICROMETER: 1e3,
    LengthUnit.MILLIMETER: 1e6,
    LengthUnit.CENTIMETER: 1e7,
    LengthUnit.DECIMETER: 1e8,
    LengthUnit.METER: 1e9,
    LengthUnit.HECTOMETER: 1e11,
    LengthUnit.KILOMETER: 1e12,
    LengthUnit.MEGAMETER: 1e15,
    LengthUnit.GIGAMETER: 1e18,
    LengthUnit.TERAMETER: 1e21,
    LengthUnit.PETAMETER: 1e24,
    LengthUnit.EXAMETER: 1e27,
    LengthUnit.ZETTAMETER: 1e30,
    LengthUnit.YOTTAMETER: 1e33,
    LengthUnit.ANGSTROM: 0.1,
    LengthUnit.INCH: 25400000.0,
    LengthUnit.FOOT: 304800000.0,
    LengthUnit.YARD: 914400000.0,
    LengthUnit.MILE: 1609344000000.0,
    LengthUnit.PARSEC: 3.085677581e25,
}


def length_unit_from_str(unit: str) -> LengthUnit:
    if unit in _STR_TO_UNIT_MAP:
        return _STR_TO_UNIT_MAP[unit]
    raise ValueError(f"Unknown unit: {unit}")



import logging
from os import PathLike
from pathlib import Path
from typing import List, Optional, TypeVar, Union, cast

import numpy as np
from rich.progress import track

from webknossos.dataset.length_unit import length_unit_from_str

from ..dataset import Dataset, LayerCategoryType
from ..dataset.properties import LayerViewConfiguration, VoxelSize, dataset_converter
from ..geometry import BoundingBox, Mag, Vec3Int
from .context import _get_context

logger = logging.getLogger(__name__)


T = TypeVar("T")


_DOWNLOAD_CHUNK_SHAPE = Vec3Int(512, 512, 512)


def download_dataset(
    dataset_name: str,
    organization_id: str,
    sharing_token: Optional[str] = None,
    bbox: Optional[BoundingBox] = None,
    layers: Optional[List[str]] = None,
    mags: Optional[List[Mag]] = None,
    path: Optional[Union[PathLike, str]] = None,
    exist_ok: bool = False,
) -> Dataset:
    context = _get_context()
    api_client = context.api_client
    api_dataset = api_client.dataset_info(
        organization_id, dataset_name, sharing_token=sharing_token
    )

    datastore_client = context.get_datastore_api_client(api_dataset.data_store.url)
    optional_datastore_token = sharing_token or context.datastore_token

    actual_path = Path(dataset_name) if path is None else Path(path)
    if actual_path.exists():
        logger.warning(f"{actual_path} already exists, skipping download.")
        return Dataset.open(actual_path)

    api_data_layers = api_dataset.data_source.data_layers
    scale = api_dataset.data_source.scale
    if api_data_layers is None or len(api_data_layers) == 0 or scale is None:
        raise RuntimeError(
            f"Could not download dataset {dataset_name}: {api_dataset.data_source.status or 'Unknown error.'}"
        )
    dataset = Dataset(
        actual_path,
        name=api_dataset.name,
        voxel_size_with_unit=VoxelSize(scale.factor, length_unit_from_str(scale.unit)),
        exist_ok=exist_ok,
    )
    for layer_name in layers or [i.name for i in api_data_layers]:
        matching_api_data_layers = [i for i in api_data_layers if i.name == layer_name]
        assert (
            len(matching_api_data_layers) > 0
        ), f"The provided layer name {layer_name} could not be found in the requested dataset."
        assert (
            len(matching_api_data_layers) == 1
        ), f"The provided layer name {layer_name} was found multiple times in the requested dataset."
        api_data_layer = matching_api_data_layers[0]
        category = cast(LayerCategoryType, api_data_layer.category)
        layer = dataset.add_layer(
            layer_name=layer_name,
            category=category,
            dtype_per_layer=api_data_layer.element_class,
            num_channels=3 if api_data_layer.element_class == "uint24" else 1,
            largest_segment_id=api_data_layer.largest_segment_id,
        )

        if api_data_layer.default_view_configuration is not None:
            default_view_configuration = dataset_converter.structure(
                api_data_layer.default_view_configuration, LayerViewConfiguration
            )
            layer.default_view_configuration = default_view_configuration

        if bbox is None:
            response_bbox = api_data_layer.bounding_box
            layer.bounding_box = BoundingBox(
                Vec3Int(response_bbox.top_left),
                Vec3Int(response_bbox.width, response_bbox.height, response_bbox.depth),
            )
        else:
            assert isinstance(
                bbox, BoundingBox
            ), f"Expected a BoundingBox object for the bbox parameter but got {type(bbox)}"
            layer.bounding_box = bbox
        if mags is None:
            mags = [Mag(mag) for mag in api_data_layer.resolutions]
        for mag in mags:
            mag_view = layer.get_or_add_mag(
                mag,
                compress=True,
                chunk_shape=Vec3Int.full(32),
                chunks_per_shard=_DOWNLOAD_CHUNK_SHAPE // 32,
            )
            aligned_bbox = layer.bounding_box.align_with_mag(mag, ceil=True)
            download_chunk_shape_in_mag = _DOWNLOAD_CHUNK_SHAPE * mag.to_vec3_int()
            for chunk in track(
                list(
                    aligned_bbox.chunk(
                        download_chunk_shape_in_mag, download_chunk_shape_in_mag
                    )
                ),
                description=f"Downloading layer={layer.name} mag={mag}",
            ):
                chunk_in_mag = chunk.in_mag(mag)
                chunk_bytes, missing_buckets = datastore_client.dataset_get_raw_data(
                    organization_name=organization_id,
                    dataset_name=dataset_name,
                    data_layer_name=layer_name,
                    mag=mag.to_long_layer_name(),
                    token=optional_datastore_token,
                    x=chunk.topleft.x,
                    y=chunk.topleft.y,
                    z=chunk.topleft.z,
                    width=chunk_in_mag.size.x,
                    height=chunk_in_mag.size.y,
                    depth=chunk_in_mag.size.z,
                )
                assert (
                    missing_buckets == "[]"
                ), f"Download contained missing buckets {missing_buckets}."
                data = np.frombuffer(
                    chunk_bytes, dtype=layer.dtype_per_channel
                ).reshape(layer.num_channels, *chunk_in_mag.size, order="F")
                mag_view.write(data, absolute_offset=chunk.topleft)
    return dataset



import tempfile
from os import PathLike
from pathlib import Path
from typing import Set, Tuple

import numpy as np
from pims import FramesSequenceND

try:
    import tifffile
except ImportError as e:
    raise ImportError(
        "Cannot import tifffile, please install it e.g. using 'webknossos[tifffile]'"
    ) from e


class PimsTiffReader(FramesSequenceND):
    @classmethod
    def class_exts(cls) -> Set[str]:
        return {"tif", "tiff"}

    # class_priority is used in pims to pick the reader with the highest priority.
    # We decided to use a custom reader for tiff files to support images with more than 3 dimensions out of the box.
    # Default is 10, and bioformats priority is 2.
    # Our custom reader for imagej_tiff has priority 20.
    # See http://soft-matter.github.io/pims/v0.6.1/custom_readers.html#plugging-into-pims-s-open-function
    class_priority = 19

    def __init__(self, path: PathLike) -> None:
        super().__init__()

        path = Path(path)

        _tiff = tifffile.TiffFile(path).series[0]
        self._tiff_axes = tuple(_tiff.axes.lower())
        for axis, shape in zip(self._tiff_axes, _tiff.shape):
            self._init_axis(axis, shape)

        self._tiff_shape = _tiff.shape

        # Selecting the first page to get the dtype and shape
        if hasattr(_tiff, "pages"):
            _tmp = _tiff.pages[0]
        else:
            _tmp = _tiff["pages"][0]  # type: ignore
        assert _tmp is not None, "No pages found in tiff file."
        self._dtype = _tmp.dtype or np.dtype("uint8")
        self._shape = _tmp.shape
        try:
            self._memmap = tifffile.memmap(path, mode="r")
        except ValueError:
            # If the tiff file is not memory-mappable, we create a temporary file and copy the data to it.
            with tempfile.TemporaryFile() as tmp:
                self._memmap = np.memmap(
                    tmp,
                    dtype=self._dtype,
                    shape=_tiff.shape,
                    mode="w+",
                )
            # The axes per page are a subset of the axes of the tifffile. We copy the data from the tifffile to the memmap with correct axes.
            # While the actual axes are e.g. ["t", "z", "y", "x"], with a shape like (3, 5, 100, 200), the axes of the tifffile consist of the axes of a singe page, e.g. ["y", "x"]. And the number of pages is 15.
            # We have to iterate over the pages and copy the data to the correct position in the memmap.
            for i in range(len(_tiff.pages)):
                other_axes = tuple(
                    axis for axis in self._tiff_axes if axis not in _tmp.axes.lower()
                )
                slices = []
                for j, axis in enumerate(other_axes):
                    size = self.sizes[axis]
                    index = (
                        i
                        // (
                            np.prod(
                                [self.sizes[axis] for axis in other_axes[j + 1 :]],
                                dtype=int,
                            )
                        )
                        % size
                    )
                    slices.append(slice(index, index + 1))

                _tiff.asarray(key=i, out=self._memmap[tuple(slices)])
        if "c" in self._tiff_axes:
            self._register_get_frame(self.get_frame_2D, "cyx")
        else:
            self._register_get_frame(self.get_frame_2D, "yx")

    def get_frame_2D(self, **ind: int) -> np.ndarray:
        # A frame of the tiff file might have less axes than the desired shape of a frame in the FramesSequenceND.
        # To get the desired axes we need to iterate over the axes of the FramesSequenceND and extract the data from the tiff file.
        slice_tuple = tuple(
            slice(None) if axis in self.bundle_axes else slice(ind[axis], ind[axis] + 1)
            for axis in self._tiff_axes
        )

        data = self._memmap[slice_tuple]
        data = np.moveaxis(
            data,
            tuple(self._tiff_axes.index(axis) for axis in self.bundle_axes),
            range(len(self.bundle_axes)),
        )
        return data.squeeze(axis=tuple(range(len(self.bundle_axes), len(data.shape))))

    @property
    def pixel_type(self) -> np.dtype:
        return self._dtype

    @property
    def shape(self) -> Tuple[int, ...]:
        return self._tiff_shape

    @property
    def frame_shape(self) -> Tuple[int, ...]:
        return self._shape



"""This module takes care of upsampling WEBKNOSSOS datasets."""

from argparse import Namespace
from multiprocessing import cpu_count
from typing import Any, Optional

import typer
from typing_extensions import Annotated

from ..dataset import Dataset, SamplingModes
from ..geometry import Mag
from ..utils import get_executor_for_args
from ._utils import DistributionStrategy, SamplingMode, parse_mag, parse_path


def main(
    *,
    source: Annotated[
        Any,
        typer.Argument(
            help="Path to your WEBKNOSSOS dataset.",
            show_default=False,
            parser=parse_path,
        ),
    ],
    sampling_mode: Annotated[
        SamplingMode, typer.Option(help="The sampling mode to use.")
    ] = SamplingMode.ANISOTROPIC,
    from_mag: Annotated[
        Mag,
        typer.Option(
            help="Mag to start upsampling from. \
Should be number or minus separated string (e.g. 2 or 2-2-2).",
            parser=parse_mag,
        ),
    ],
    layer_name: Annotated[
        Optional[str],
        typer.Option(
            help="Name of the layer that should be downsampled.", show_default=False
        ),
    ] = None,
    jobs: Annotated[
        int,
        typer.Option(
            help="Number of processes to be spawned.",
            rich_help_panel="Executor options",
        ),
    ] = cpu_count(),
    distribution_strategy: Annotated[
        DistributionStrategy,
        typer.Option(
            help="Strategy to distribute the task across CPUs or nodes.",
            rich_help_panel="Executor options",
        ),
    ] = DistributionStrategy.MULTIPROCESSING,
    job_resources: Annotated[
        Optional[str],
        typer.Option(
            help='Necessary when using slurm as distribution strategy. Should be a JSON string \
(e.g., --job_resources=\'{"mem": "10M"}\')\'',
            rich_help_panel="Executor options",
        ),
    ] = None,
) -> None:
    """Upsample your WEBKNOSSOS dataset."""

    executor_args = Namespace(
        jobs=jobs,
        distribution_strategy=distribution_strategy.value,
        job_resources=job_resources,
    )
    dataset = Dataset.open(source)
    mode = SamplingModes.parse(sampling_mode.value)

    if layer_name is None:
        upsample_all_layers(dataset, mode, from_mag, executor_args)
    else:
        with get_executor_for_args(args=executor_args) as executor:
            layer = dataset.get_layer(layer_name)
            layer.upsample(from_mag=from_mag, sampling_mode=mode, executor=executor)


def upsample_all_layers(
    dataset: Dataset, mode: SamplingModes, from_mag: Mag, executor_args: Namespace
) -> None:
    """Iterates over all layers and upsamples them."""

    for layer in dataset.layers.values():
        with get_executor_for_args(args=executor_args) as executor:
            layer.upsample(
                from_mag=from_mag,
                sampling_mode=mode,
                executor=executor,
            )



"""This module converts a Zarr dataset to a WEBKNOSSOS dataset."""

import logging
import time
from argparse import Namespace
from functools import partial
from multiprocessing import cpu_count
from pathlib import Path
from typing import Any, Optional, Tuple, Union, cast

import numpy as np
import typer
import zarr
from typing_extensions import Annotated

from webknossos.dataset.length_unit import LengthUnit
from webknossos.dataset.properties import DEFAULT_LENGTH_UNIT_STR, VoxelSize

from ..dataset import DataFormat, Dataset, MagView, SegmentationLayer
from ..dataset._array import _fsstore_from_path
from ..dataset.defaults import DEFAULT_CHUNK_SHAPE, DEFAULT_CHUNKS_PER_SHARD
from ..geometry import BoundingBox, Mag, Vec3Int
from ..utils import get_executor_for_args, wait_and_ensure_success
from ._utils import (
    DistributionStrategy,
    SamplingMode,
    VoxelSizeTuple,
    parse_mag,
    parse_path,
    parse_vec3int,
    parse_voxel_size,
)

logger = logging.getLogger(__name__)


def _zarr_chunk_converter(
    bounding_box: BoundingBox,
    source_zarr_path: Path,
    target_mag_view: MagView,
    flip_axes: Optional[Union[int, Tuple[int, ...]]],
) -> int:
    logging.info("Conversion of %s", bounding_box.topleft)

    slices = bounding_box.to_slices()
    zarr_file = zarr.open(store=_fsstore_from_path(source_zarr_path), mode="r")
    source_data: Any = zarr_file[slices][None, ...]

    if flip_axes:
        source_data = np.flip(source_data, flip_axes)

    contiguous_chunk = source_data.copy(order="F")
    target_mag_view.write(contiguous_chunk, bounding_box.topleft)

    return source_data.max()


def convert_zarr(
    source_zarr_path: Path,
    target_path: Path,
    layer_name: str,
    data_format: DataFormat,
    chunk_shape: Vec3Int,
    chunks_per_shard: Vec3Int,
    is_segmentation_layer: bool = False,
    voxel_size_with_unit: VoxelSize = VoxelSize((1.0, 1.0, 1.0)),
    flip_axes: Optional[Union[int, Tuple[int, ...]]] = None,
    compress: bool = True,
    executor_args: Optional[Namespace] = None,
) -> MagView:
    """Performs the conversation of a Zarr dataset to a WEBKNOSSOS dataset."""
    ref_time = time.time()

    file = zarr.open(store=_fsstore_from_path(source_zarr_path), mode="r")
    input_dtype: Any = file.dtype
    shape: Any = file.shape

    wk_ds = Dataset(
        target_path, voxel_size_with_unit=voxel_size_with_unit, exist_ok=True
    )
    wk_layer = wk_ds.get_or_add_layer(
        layer_name,
        "segmentation" if is_segmentation_layer else "color",
        dtype_per_layer=np.dtype(input_dtype),
        num_channels=1,
        largest_segment_id=0,
        data_format=data_format,
    )
    wk_layer.bounding_box = BoundingBox((0, 0, 0), shape)
    wk_mag = wk_layer.get_or_add_mag(
        "1",
        chunk_shape=chunk_shape,
        chunks_per_shard=chunks_per_shard,
        compress=compress,
    )

    # Parallel chunk conversion
    with get_executor_for_args(args=executor_args) as executor:
        largest_segment_id_per_chunk = wait_and_ensure_success(
            executor.map_to_futures(
                partial(
                    _zarr_chunk_converter,
                    source_zarr_path=source_zarr_path,
                    target_mag_view=wk_mag,
                    flip_axes=flip_axes,
                ),
                wk_layer.bounding_box.chunk(chunk_shape=chunk_shape * chunks_per_shard),
            ),
            executor=executor,
        )

    if is_segmentation_layer:
        largest_segment_id = int(max(largest_segment_id_per_chunk))
        cast(SegmentationLayer, wk_layer).largest_segment_id = largest_segment_id

    logger.debug(
        "Conversion of %s took %.8fs", source_zarr_path, time.time() - ref_time
    )
    return wk_mag


def main(
    *,
    source: Annotated[
        Any,
        typer.Argument(
            help="Path to your raw image data.",
            show_default=False,
            parser=parse_path,
        ),
    ],
    target: Annotated[
        Any,
        typer.Argument(
            help="Target path to save your WEBKNOSSOS dataset.",
            show_default=False,
            parser=parse_path,
        ),
    ],
    layer_name: Annotated[
        str,
        typer.Option(help="Name of the cubed layer (color or segmentation)"),
    ] = "color",
    voxel_size: Annotated[
        VoxelSizeTuple,
        typer.Option(
            help="The size of one voxel in source data in nanometers. "
            "Should be a comma separated string (e.g. 11.0,11.0,20.0).",
            parser=parse_voxel_size,
            metavar="VoxelSize",
            show_default=False,
        ),
    ],
    unit: Annotated[
        LengthUnit,
        typer.Option(
            help="The unit of the voxel size.",
        ),
    ] = DEFAULT_LENGTH_UNIT_STR,  # type:ignore
    data_format: Annotated[
        DataFormat,
        typer.Option(
            help="Data format to store the target dataset in.",
        ),
    ] = "wkw",  # type:ignore
    chunk_shape: Annotated[
        Vec3Int,
        typer.Option(
            help="Number of voxels to be stored as a chunk in the output format "
            "(e.g. `32` or `32,32,32`).",
            parser=parse_vec3int,
            metavar="Vec3Int",
        ),
    ] = DEFAULT_CHUNK_SHAPE,
    chunks_per_shard: Annotated[
        Vec3Int,
        typer.Option(
            help="Number of chunks to be stored as a shard in the output format "
            "(e.g. `32` or `32,32,32`).",
            parser=parse_vec3int,
            metavar="Vec3Int",
        ),
    ] = DEFAULT_CHUNKS_PER_SHARD,
    max_mag: Annotated[
        Optional[Mag],
        typer.Option(
            help="Max resolution to be downsampled. "
            "Should be number or minus separated string (e.g. 2 or 2-2-2).",
            parser=parse_mag,
        ),
    ] = None,
    interpolation_mode: Annotated[
        str,
        typer.Option(
            help="The interpolation mode that should be used "
            "(median, mode, nearest, bilinear or bicubic)."
        ),
    ] = "default",
    flip_axes: Annotated[
        Optional[Vec3Int],
        typer.Option(
            help="The axes at which should be flipped. "
            "Input format is a comma separated list of axis indices. "
            "For example, 1,2,3 will flip the x, y and z axes.",
            parser=parse_vec3int,
            metavar="Vec3Int",
        ),
    ] = None,
    compress: Annotated[
        bool, typer.Option(help="Enable compression of the target dataset.")
    ] = False,
    sampling_mode: Annotated[
        SamplingMode, typer.Option(help="The sampling mode to use.")
    ] = SamplingMode.ANISOTROPIC,
    is_segmentation_layer: Annotated[
        bool,
        typer.Option(
            help="When converting one layer, signals whether layer is segmentation layer. \
When converting a folder, this option is ignored."
        ),
    ] = False,
    jobs: Annotated[
        int,
        typer.Option(
            help="Number of processes to be spawned.",
            rich_help_panel="Executor options",
        ),
    ] = cpu_count(),
    distribution_strategy: Annotated[
        DistributionStrategy,
        typer.Option(
            help="Strategy to distribute the task across CPUs or nodes.",
            rich_help_panel="Executor options",
        ),
    ] = DistributionStrategy.MULTIPROCESSING,
    job_resources: Annotated[
        Optional[str],
        typer.Option(
            help='Necessary when using slurm as distribution strategy. Should be a JSON string \
(e.g., --job_resources=\'{"mem": "10M"}\')\'',
            rich_help_panel="Executor options",
        ),
    ] = None,
) -> None:
    """Converts a Zarr dataset to a WEBKNOSSOS dataset."""

    if not source.is_dir():
        logger.error("source_path is not a directory")
        return

    executor_args = Namespace(
        jobs=jobs,
        distribution_strategy=distribution_strategy.value,
        job_resources=job_resources,
    )
    voxel_size_with_unit = VoxelSize(voxel_size, unit)

    mag_view = convert_zarr(
        source,
        target,
        layer_name=layer_name,
        data_format=data_format,
        chunk_shape=chunk_shape,
        chunks_per_shard=chunks_per_shard,
        is_segmentation_layer=is_segmentation_layer,
        voxel_size_with_unit=voxel_size_with_unit,
        flip_axes=flip_axes,
        compress=compress,
        executor_args=executor_args,
    )

    with get_executor_for_args(executor_args) as executor:
        mag_view.layer.downsample(
            from_mag=mag_view.mag,
            coarsest_mag=max_mag,
            interpolation_mode=interpolation_mode,
            compress=compress,
            sampling_mode=sampling_mode,
            executor=executor,
        )
